package nrs.findclass;

import java.io.*;
import java.util.*;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

/*
 * This file is part of FindClass.
 * FindClass is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * FindClass is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FindClass; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * Find classes or other files in JARs, Zips or directories.
 * <p>See http://www.nsydenham.net/java/FindClass/FindClass.shtml for details.</p>
 *
 * @author Copyright (C) 2003-2004 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class FindClass
{
	private List files = new ArrayList();
	private List locations = new ArrayList();
	private List matches = new ArrayList();
	private boolean caseSensitive = true;
	private boolean exactMatch = false;
	private boolean recurse = false;
	private boolean matchArchives = false;
	private boolean verbose = true;
	private PrintStream print = System.out;
	private final static String fs = System.getProperty("path.separator");
	public final static String version = "1.2";
	
	// constants
	private final static String JAR = "jar";
	private final static String ZIP = "zip";
		
	/** Filters for jar, zip and class files plus directories */
	class JavaFilter implements FileFilter
	{
		public boolean accept(File file)
		{
			if (file.isDirectory()) return true;
			else
			{
				String lName = file.getName().toLowerCase();
				if (lName.endsWith(JAR) || lName.endsWith(ZIP)) {
					return true;
				}
			}
			
			return false;
		}
	}
	
	/** Iterate through locations looking for files */
	public void findFiles()
	{
		if (files.size() == 0)
		{
			print.println("Please specify one or more files to search for.");
			printHelp();
		}
		
		// add system classpath
		if (locations.size() == 0)
		{
			addLocations(System.getProperty("java.class.path"));
		}
		
		if (locations.size() > 0)
		{
			searchFiles();
		}
		
		log("Finished search. Found "+matches.size()+" matches.");
		
		if (verbose && matches.size() > 0)
		{
			log("\nMatches:");
			Collections.sort(matches);
			Iterator it = matches.iterator();
			while (it.hasNext())
			{
				log((String)it.next());
			}
		}
		print.close();
	}
	
	/**
	 * Add a file to look for
	 * @param fileName file to look for.
	 */
	public void addFile(String fileName)
	{
		files.add(fileName);
	}
	
	/** Search for files in the path */
	private void searchFiles()
	{
		String location;
		String fileName;
		Iterator fit, lit;
		
		// print search locations
		if (verbose)
		{
			print.println("Search path: ");
			Iterator it = locations.iterator();
			while (it.hasNext())
			{
				print.println("    "+it.next());
			}
		}
		
		fit = files.iterator();
		while (fit.hasNext())
		{
			fileName = (String)fit.next();
			log("Searching for "+fileName);
			
			lit = locations.iterator();
			while (lit.hasNext())
			{
				location = (String)lit.next();
				
				if (location.toLowerCase().endsWith(JAR) || location.toLowerCase().endsWith(ZIP)) {
					findInArchive(fileName, location);
				} else {
					findInDirectory(fileName, location);
				}
			}
		}
	}
	
	/**
	 * Check if the file is part of the specified path based on case and exact match parameters
	 * @param fileName the name being searched for
	 * @param path the path being searched
	 */
	private boolean isMatch(String fileName, String path)
	{
		if (exactMatch &&
			(caseSensitive && path.equals(fileName)) || 
			(!caseSensitive && path.equalsIgnoreCase(fileName)))
		{
			return true;
		}
		else if (!exactMatch &&
			(caseSensitive && path.indexOf(fileName) > -1) ||
			(!caseSensitive && path.toLowerCase().indexOf(fileName.toLowerCase()) > -1))
		{
			return true;
		}
		
		return false;
	}
	
	/**
	 * See if the file exists in the specified JAR or ZIP file
	 * @param fileName name of the file to look for
	 * @param archive name of the archive to examine
	 */
	private void findInArchive(String fileName, String archive)
	{
		ZipEntry ze;
		String name;
		
		log("Searching "+archive);
		
		try
		{
			// use a JarFile because it can also read zips
			JarFile jf = new JarFile(archive);
			Enumeration enum = jf.entries();
			while (enum.hasMoreElements())
			{
				ze = (ZipEntry)enum.nextElement();
				name = ze.getName();
				if (isMatch(fileName, name)) {
					print.println("    FOUND "+fileName+" in "+archive);
					matches.add(fileName+" in "+archive);
				}
			}
		}
		catch (FileNotFoundException fnex)
		{
			print.println("    File "+archive+" doesn't exist");
		}
		catch (IOException ioex)
		{
			print.println("    Error opening file "+archive);
		}
	}
	
	/**
	 * See if the file exists in the specified directory
	 * @param fileName name of the file to look for
	 * @parm dir name of the directory to search
	 */
	private void findInDirectory(String fileName, String dir)
	{
		File dirFile = new File(dir);
		File dirEntries[];
		File dirEntry;
		String name;
		
		log("Searching "+dir);
		
		// check directory is valid
		if (!dirFile.exists())
		{
			print.println("    "+dir+" doesn't exist");
			return;
		}
		else if (!dirFile.isDirectory())
		{
			print.println("    "+dir+" is not a directory");
			return;
		}
		
		if (matchArchives) {
			dirEntries = dirFile.listFiles(new JavaFilter());
		} else {
			dirEntries = dirFile.listFiles();
		}
		
		for (int i=0; i<dirEntries.length; i++)
		{
			dirEntry = dirEntries[i];
			name = dirEntry.getName();

			if (recurse && dirEntry.isDirectory())
			{
				findInDirectory(fileName, dirEntry.getAbsolutePath());
			}
			else if (name.endsWith(JAR) || name.endsWith(ZIP))
			{
				findInArchive(fileName, dirEntry.getAbsolutePath());
			}
			else if (isMatch(fileName, name))
			{
				print.println("    FOUND "+fileName+" in "+dir);
				matches.add(fileName+" in "+dir);
			}
		}
	}
	
	/**
	 * Add entries
	 * @param filePath the path to search
	 */
	private void addLocations(String filePath)
	{
		StringTokenizer st = new StringTokenizer(filePath, fs);
		String cpEntry;
		
		while (st.hasMoreTokens())
		{
			locations.add(st.nextToken());
		}
	}
	
	/**
	 * Set whether the tool should be verbose or not
	 * @param verbose if true print progress
	 */
	public void setVerbose(boolean verbose)
	{
		this.verbose = verbose;
	}
	
	/**
	 * Uses the verbose variable to determine if the message should be displayed
	 * @param message to display
	 */
	private void log(String message)
	{
		if (verbose) print.println(message);
	}
	
	/**
	 * If set to true, recurse into directories
	 * @param recurse 
	 */
	public void setRecurse(boolean recurse)
	{
		this.recurse = recurse;
	}
	
	/**
	 * Set the OutputStream to something other than System.out
	 * @param outs the new OutputStream
	 */
	public void setOutputStream(OutputStream outs)
	{
		print = new PrintStream(outs);
	}
	
	/**
	 * Set the output to the specified file
	 * @param filename file to log to
	 * @exception FileNotFoundException if the specified file doesn't exist
	 */
	public void setOutputFile(String filename) throws FileNotFoundException
	{
		if (filename == null) return;
		setOutputStream(new FileOutputStream(filename));
	}
	
	/**
	 * Set whether the search is case sensitive
	 * @param cs if true the search will be case sensitive
	 */
	public void setCaseSensitive(boolean cs) {
		this.caseSensitive = cs;
	}
	
	/**
	 * Set whether matching will be exact to the word
	 * @param exactMatch if true match search param exactly
	 */
	public void setExactMatch(boolean exactMatch) {
		this.exactMatch = exactMatch;
	}
	
	public void setArchiveMatch(boolean archiveMatch) {
		this.matchArchives = archiveMatch;
	}
	
	/** Print a help message */
	private static void printHelp()
	{
		printAbout();
		
		System.out.println("java nrs.findclass.FindClass <options> <classes>");
		System.out.println("Do _not_ specify packages.");
		System.out.println("If no classpath or directory is specified the system classpath will be searched.");
		System.out.println("Multiple files can be searched for at the same time.");
		System.out.println("	-a		print info about this program");
		System.out.println("	-c <classpath>	add specified classpath to system classpath");
		System.out.println("	-d <path>	search only in specified path(s)");
		System.out.println("	-f <file>	log output to file");
		System.out.println("	-h		this help message");
		System.out.println("	-i		ignore case (default: no)");
		System.out.println("	-j		only search Java archives (jar, zip) (default: no)");
		System.out.println("	-m		match word (default: no)");
		System.out.println("	-q		only print when a match is found\n");
		System.exit(0);
	}
	
	/** Print some info about the utility */
	private static void printAbout()
	{
		System.out.println("FindClass "+version);
		System.out.println("Searches a classpath for the specified file(s).");
		System.out.println("This utility is distributed under the terms of the GPL and comes with absolutely no warranty");
		System.out.println("Homepage: http://www.nsydenham.net/java/FindClass/FindClass.shtml");
		System.out.println("Copyright (C) 2002-2004 Nick Sydenham <nsydenham@yahoo.co.uk>\n");
	}
	
	/** Start the utility */
	public static void main(String[] args)
	{
		String arg;
		
		FindClass find = new FindClass();
		
		// parse the command line arguments
		try
		{
			for (int i=0; i<args.length; i++)
			{
				arg = args[i];
				if (arg.equalsIgnoreCase("-a"))
				{
					printAbout();
					System.exit(0);
				}
				else if (arg.equals("-c"))
				{
					// add locations in class path
					find.addLocations(System.getProperty("java.class.path"));
					// add extra locations
					find.addLocations(args[++i]);
				}
				else if (arg.equals("-d"))
				{
					// search only in specifed locations
					find.addLocations(args[++i]);
				}
				else if (arg.equalsIgnoreCase("-f"))
				{
					find.setOutputFile(args[++i]);
				}
				else if (arg.equalsIgnoreCase("-h"))
				{
					printHelp();
				}
				else if (arg.equalsIgnoreCase("-i"))
				{
					find.setCaseSensitive(false);
				}
				else if (arg.equalsIgnoreCase("-j"))
				{
					find.setArchiveMatch(true);
				}
				else if (arg.equalsIgnoreCase("-m"))
				{
					find.setExactMatch(true);
				}
				else if (arg.equalsIgnoreCase("-q"))
				{
					find.setVerbose(false);
				}
				else if (arg.equalsIgnoreCase("-r"))
				{
					find.setRecurse(true);
				}
				else
				{
					find.addFile(arg);
				}
			}
		} catch (IndexOutOfBoundsException iex)	{
			System.out.println("Invalid arguments");
			printHelp();
		} catch (FileNotFoundException fnex) {
			System.out.println(fnex.getMessage());
			System.exit(-1);
		}
		
		find.findFiles();
	}
}
